package com.css.simulation.resource.scheduler.consumer;


import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.CollectionUtil;
import api.common.util.JsonUtil;
import api.common.util.StringUtil;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.*;
import com.css.simulation.resource.scheduler.pojo.to.PrefixTO;
import com.css.simulation.resource.scheduler.service.ProjectService;
import com.css.simulation.resource.scheduler.util.ProjectUtil;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Component;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

@Component
@Slf4j
public class ProjectConsumer {

    @Value("${scheduler.minio-path.project-result}")
    String resultPathMinio;
    @Value("${scheduler.linux-path.job-template}")
    String jobTemplate;
    @Value("${scheduler.linux-path.job-yaml}")
    String jobYaml;

    @Value("${scheduler.host.hostname}")
    String hostname;
    @Value("${scheduler.host.username}")
    String username;
    @Value("${scheduler.host.password}")
    String password;

    // -------------------------------- Comment --------------------------------

    @Autowired
    KafkaTemplate<String, String> kafkaTemplate;
    @Autowired
    StringRedisTemplate redisTemplate;
    @Autowired
    ManualProjectMapper manualProjectMapper;
    @Autowired
    AutoProjectMapper autoProjectMapper;
    @Autowired
    AutoSubProjectMapper autoSubProjectMapper;
    @Autowired
    TaskMapper taskMapper;
    @Autowired
    IndexMapper indexMapper;
    @Autowired
    IndexTemplateMapper indexTemplateMapper;
    @Autowired
    SceneMapper sceneMapper;
    @Autowired
    VehicleMapper vehicleMapper;
    @Autowired
    SensorCameraMapper sensorCameraMapper;
    @Autowired
    SensorOgtMapper sensorOgtMapper;
    @Autowired
    AlgorithmMapper algorithmMapper;
    @Autowired
    UserMapper userMapper;
    @Autowired
    ClusterMapper clusterMapper;
    @Autowired
    ProjectService projectService;
    @Autowired
    ProjectUtil projectUtil;


    /**
     * 任务运行前首先判断用户是否拥有可分配资源
     *
     * @param projectRecord 项目启动消息
     */
    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.start-topic}")
    @SneakyThrows
    public void cacheProject(ConsumerRecord<String, String> projectRecord) {
        String projectJson = projectRecord.value();
        log.info("ProjectConsumer--cacheManualProject 接收到项目开始消息为：" + projectJson);
        //1 读取 kafka 的 project 信息
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(projectJson, ProjectMessageDTO.class);
        String projectId = projectMessageDTO.getProjectId();    // 手动执行项目 id 或 自动执行子项目 id
        Long parallelism = projectMessageDTO.getParallelism();   // 项目并行度
        String projectType = projectMessageDTO.getType(); // 项目类型
        //2 根据 projectId 获取创建用户 id
        String userId;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userId = manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userId = autoSubProjectMapper.selectCreateUserById(projectId);
        } else {
            log.error("ProjectConsumer--cacheManualProject 项目类型错误：" + projectJson);
            return;
        }
        if (StringUtil.isEmpty(userId)) {
            log.error("ProjectConsumer--cacheManualProject 未查询到项目创建人：" + projectJson);
            return;
        }
        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserPO userPO = userMapper.selectById(userId);
        String roleCode = userPO.getRoleCode();
        String useType = userPO.getUseType();
        ClusterPO clusterPO;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(DictConstants.SYSTEM_CLUSTER_ID, projectId);
            run(DictConstants.SYSTEM_CLUSTER_ID, projectId, redisPrefix.getProjectRunningKey(), projectJson);
            return;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterPO = clusterMapper.selectByUserId(userId);
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USE_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterPO = clusterMapper.selectByUserId(userId);
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userPO.getCreateUserId();
                clusterPO = clusterMapper.selectByUserId(parentUserId);
            }
        } else {
            log.error("ProjectConsumer--cacheManualProject 未知账户类型，不予执行：" + projectJson);
            return;
        }
        // 获取拥有的节点数量，即仿真软件证书数量
        String clusterId = clusterPO.getId();
        int simulationLicenseNumber = clusterPO.getNumSimulationLicense();
        // 获取该集群中正在运行的项目，如果没有则立即执行
        PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(clusterId, projectId);
        Set<String> clusterRunningKeySet = redisTemplate.keys(redisPrefix.getClusterRunningPrefix() + "*");
        List<String> runningProjectSet;
        if (CollectionUtil.isEmpty(clusterRunningKeySet)) {
            run(clusterId, projectId, redisPrefix.getProjectRunningKey(), projectJson);
            return;
        }
        runningProjectSet = projectUtil.getRunningProjectList(clusterRunningKeySet);
        if (CollectionUtil.isEmpty(runningProjectSet)) {
            run(clusterId, projectId, redisPrefix.getProjectRunningKey(), projectJson);
            return;
        }
        // 计算正在运行的项目的并行度总和
        long parallelismSum = 0;
        for (String projectKey : runningProjectSet) {
            String projectJsonTemp = redisTemplate.opsForValue().get(projectKey);
            ProjectMessageDTO projectMessageTemp = JsonUtil.jsonToBean(projectJsonTemp, ProjectMessageDTO.class);
            parallelismSum += projectMessageTemp.getParallelism();
        }
        // 如果执行后的并行度总和小于最大节点数则执行，否则不执行
        if (parallelismSum + parallelism <= simulationLicenseNumber) {
            run(clusterId, projectId, redisPrefix.getProjectRunningKey(), projectJson);
        } else {
            wait(clusterId, projectId, redisPrefix.getProjectWaitingKey(), projectJson);
        }
    }

    public void run(String clusterId, String projectId, String projectRunningKey, String projectJson) {
        log.info("ProjectConsumer--run 集群 " + clusterId + " 将项目 " + projectId + " 执行！");
        redisTemplate.opsForValue().set(projectRunningKey, projectJson);
        parseProject(projectJson, "cluster:" + clusterId, projectRunningKey);
    }

    public void wait(String clusterId, String projectId, String projectWaitingKey, String projectJson) {
        log.info("ProjectConsumer--cacheManualProject 集群 " + clusterId + " 将项目 " + projectId + " 放入等待队列！");
        redisTemplate.opsForValue().set(projectWaitingKey, projectJson);
    }


    /**
     * 开始执行以及重新执行
     *
     * @param projectJson 项目启动消息
     */
    @SneakyThrows
    public void parseProject(String projectJson, String clusterPrefix, String projectRunningPrefix) {

        // -------------------------------- 0 准备 --------------------------------
        log.info("ProjectConsumer--parseManualProject 接收到项目开始消息为：" + projectJson);
        //1 读取 kafka 的 project 信息
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(projectJson, ProjectMessageDTO.class);
        String projectId = projectMessageDTO.getProjectId();    // 项目 id
        String projectType = projectMessageDTO.getType();    // 项目 id
        String packageId = projectMessageDTO.getScenePackageId();   // 场景测试包 id
        Long maxSimulationTime = projectMessageDTO.getMaxSimulationTime(); // 最大仿真时间，即生成视频的时间长度
        String vehicleConfigId = projectMessageDTO.getVehicleConfigId();// 模型配置 id
        String algorithmId = projectMessageDTO.getAlgorithmId();    // 算法 id
        String userId = manualProjectMapper.selectCreateUserById(projectId);
        Long parallelism = projectMessageDTO.getParallelism();    // 并行度
        //2 执行前准备，删除改项目下所有任务，即重新执行改项目时需要新的测试包
        projectService.prepare(clusterPrefix, projectId, projectJson);
        // -------------------------------- 1 查询场景 --------------------------------
        //1-1 根据场景测试包 packageId，拿到场景集合（不包括重复场景），重复场景会在发送消息时根据叶子指标发送多次。
        List<ScenePO> scenePOList = projectService.handlePackage(projectRunningPrefix, projectId, packageId);
        Set<ScenePO> scenePOSet = new HashSet<>(scenePOList); // 如果不去重的话会出现多个场景重复关联多个指标
        // -------------------------------- 2 查询模型 --------------------------------
        //2-1 根据车辆配置id vehicleConfigId, 获取 模型信息和传感器信息
        VehiclePO vehiclePO = vehicleMapper.selectByVehicleConfigId(vehicleConfigId);   // 车辆
        List<CameraPO> cameraPOList = sensorCameraMapper.selectCameraByVehicleConfigId(vehicleConfigId);    // 摄像头
        List<OgtPO> ogtPOList = sensorOgtMapper.selectOgtByVehicleId(vehicleConfigId); // 完美传感器
        // -------------------------------- 3 发送任务消息 --------------------------------
        projectService.sendTaskMessage(projectRunningPrefix, userId, projectId, maxSimulationTime, scenePOSet, vehiclePO, cameraPOList, ogtPOList);
        // -------------------------------- 4 算法导入（一期按单机版做） --------------------------------
        String algorithmDockerImage = projectService.handleAlgorithm(projectId, algorithmId);
        // -------------------------------- 5 创建 pod 开始执行 --------------------------------
        projectService.transferAndRunYaml(
                jobTemplate + "job-template.yaml",
                projectId,
                algorithmDockerImage,
                scenePOList.size(),
                parallelism,
                jobYaml + "project-" + projectId + ".yaml"
        );

    }

    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.stop-topic}")
    @SneakyThrows
    public void stopProject(ConsumerRecord<String, String> stopRecord) {
        log.info("ProjectConsumer--stopManualProject 接收到的项目终止消息为：" + stopRecord);
        //1 读取 kafka 的项目停止信息
        /*
            {
                "projectId": "sadfasdfs",	// 项目 id
                "type": "1",	// 项目 id
            }
         */

        String json = stopRecord.value();
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(json);
        String projectId = jsonNode.path("projectId").asText();
        String type = jsonNode.path("type").asText();
        projectService.stopProject(projectId, type);
    }


}
