package com.css.simulation.resource.scheduler.util;

import api.common.util.SshUtil;
import io.kubernetes.client.openapi.ApiClient;
import io.kubernetes.client.openapi.ApiException;
import io.kubernetes.client.openapi.apis.BatchV1Api;
import io.kubernetes.client.openapi.apis.CoreV1Api;
import io.kubernetes.client.openapi.models.V1JobList;
import io.kubernetes.client.openapi.models.V1Namespace;
import io.kubernetes.client.openapi.models.V1ObjectMeta;
import io.kubernetes.client.openapi.models.V1PodList;
import lombok.extern.slf4j.Slf4j;
import org.apache.sshd.client.SshClient;
import org.apache.sshd.client.session.ClientSession;

import java.io.IOException;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

@Slf4j
public class KubernetesUtil {

    public static void applyYaml(String hostname, String username, String password, String jobTemplateYamlPathTarget) throws IOException {
        SshClient client = SshUtil.getClient();
        ClientSession session = SshUtil.getSession(client, hostname, username, password);
        SshUtil.execute(session, "kubectl apply -f " + jobTemplateYamlPathTarget);
        session.close();
        client.stop();
    }

    /**
     * 创建 namespace 命名空间
     *
     * @param apiClient     客户端
     * @param namespaceName 命名空间名称
     * @return 命名空间对象
     * @throws ApiException 异常
     */
    public static V1Namespace createNamespace(ApiClient apiClient, String namespaceName) throws ApiException {
        CoreV1Api coreV1Api = new CoreV1Api(apiClient);
        V1Namespace yaml = new V1Namespace();
        //1 apiVersion
        yaml.setApiVersion("v1");

        //2 kind
        yaml.setKind("Namespace");

        //3 metadata
        V1ObjectMeta metadata = new V1ObjectMeta();
        metadata.setName(namespaceName);
        metadata.setLabels(null);
        yaml.setMetadata(metadata);

        return coreV1Api.createNamespace(yaml, null, null, null, null);
    }

    public static List<String> getPod(ApiClient apiClient, String namespaceName) throws ApiException {

        CoreV1Api api = new CoreV1Api(apiClient);
        V1PodList list;
        if (namespaceName == null || "".equals(namespaceName) || "default".equals(namespaceName)) {
            list = api.listNamespacedPod(
                    "default",
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        } else if ("all".equals(namespaceName)) {
            list = api.listPodForAllNamespaces(
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        } else {
            list = api.listNamespacedPod(
                    namespaceName,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        }
        return list.getItems().stream().map(pod -> Objects.requireNonNull(pod.getMetadata()).getName()).collect(Collectors.toList());
    }


    /**
     * 删除 pod
     *
     * @param apiClient     api 客户端
     * @param namespaceName namespace 名称
     * @param podName       pod 名称
     * @throws ApiException 异常
     */
    public static void deletePod(ApiClient apiClient, String namespaceName, String podName) throws ApiException {
        log.info("KubernetesUtil--deletePod 删除 " + namespaceName + ":" + podName);
        CoreV1Api coreV1Api = new CoreV1Api(apiClient);
        coreV1Api.deleteNamespacedPod(podName, namespaceName, null, null, null, null, null, null);
    }

    public static List<String> getJob(ApiClient apiClient, String namespaceName) throws ApiException {
        BatchV1Api batchV1Api = new BatchV1Api(apiClient);
        V1JobList v1JobList;
        if (namespaceName == null || "".equals(namespaceName) || "default".equals(namespaceName)) {
            v1JobList = batchV1Api.listNamespacedJob(
                    "default",
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        } else if ("all".equals(namespaceName)) {
            v1JobList = batchV1Api.listJobForAllNamespaces(
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        } else {
            v1JobList = batchV1Api.listNamespacedJob(
                    namespaceName,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null,
                    null);
        }
        return v1JobList.getItems().stream().map(pod -> Objects.requireNonNull(pod.getMetadata()).getName()).collect(Collectors.toList());
    }

    public static void deleteJob(ApiClient apiClient, String namespaceName, String jobNameToDelete) throws ApiException {
        List<String> runningJobNameList = getJob(apiClient, namespaceName);
        for (String runningJobName : runningJobNameList) {
            if (jobNameToDelete.equals(runningJobName)) {
                BatchV1Api batchV1Api = new BatchV1Api(apiClient);
                batchV1Api.deleteNamespacedJob(jobNameToDelete, namespaceName, null, null, null, null, null, null);
                break;
            }
        }
    }
}
